@@ -14,13 +14,17 @@ See the License for the specific language governing permissions and
 limitations under the License.
 ]]--
 
--- This is main lua config file for rspamd
+-- Misc rules
 
 local E = {}
 local fun = require "fun"
 local util = require "rspamd_util"
+local rspamd_parsers = require "rspamd_parsers"
 local rspamd_regexp = require "rspamd_regexp"
 local rspamd_lua_utils = require "lua_util"
+local bit = require "bit"
+local rspamd_url = require "rspamd_url"
+local url_flags_tab = rspamd_url.flags
 
 -- Different text parts
 rspamd_config.R_PARTS_DIFFER = {
@@ -55,63 +59,85 @@ rspamd_config.R_PARTS_DIFFER = {
 }
 
 -- Date issues
-rspamd_config.MISSING_DATE = {
+local date_id = rspamd_config:register_symbol({
+  name = 'DATE_CB',
+  type = 'callback,mime',
   callback = function(task)
-    local date = task:get_header_raw('Date')
-    if date == nil or date == '' then
-      return true
+    local date_time = task:get_header('Date')
+    if date_time == nil or date_time == '' then
+      task:insert_result('MISSING_DATE', 1.0)
+      return
     end
-    return false
-  end,
+
+    local dm, err = rspamd_parsers.parse_smtp_date(date_time)
+    if err then
+      task:insert_result('INVALID_DATE', 1.0)
+      return
+    end
+
+    local dt = task:get_date({format = 'connect', gmt = true})
+    local date_diff = dt - dm
+
+    if date_diff > 86400 then
+      -- Older than a day
+      task:insert_result('DATE_IN_PAST', 1.0, tostring(math.floor(date_diff/3600)))
+    elseif -date_diff > 7200 then
+      -- More than 2 hours in the future
+      task:insert_result('DATE_IN_FUTURE', 1.0, tostring(math.floor(-date_diff/3600)))
+    end
+  end
+})
+
+rspamd_config:register_symbol({
+  name = 'MISSING_DATE',
   score = 1.0,
   description = 'Message date is missing',
   group = 'headers',
-  type = 'mime',
-}
+  type = 'virtual',
+  parent = date_id,
+})
 
-rspamd_config.DATE_IN_FUTURE = {
-  callback = function(task)
-    local dm = task:get_date{format = 'message', gmt = true}
-    local dt = task:get_date{format = 'connect', gmt = true}
-    -- 2 hours
-    if dm > 0 and dm - dt > 7200 then
-      return true
-    end
-    return false
-  end,
+rspamd_config:register_symbol({
+  name = 'INVALID_DATE',
+  score = 1.5,
+  description = 'Malformed date header',
+  group = 'headers',
+  type = 'virtual',
+  parent = date_id,
+})
+
+rspamd_config:register_symbol({
+  name = 'DATE_IN_FUTURE',
   score = 4.0,
   description = 'Message date is in future',
   group = 'headers',
-  type = 'mime',
-}
+  type = 'virtual',
+  parent = date_id,
+})
 
-rspamd_config.DATE_IN_PAST = {
-  callback = function(task)
-    local dm = task:get_date{format = 'message', gmt = true}
-    local dt = task:get_date{format = 'connect', gmt = true}
-    -- A day
-    if dm > 0 and dt - dm > 86400 then
-      return true
-    end
-    return false
-  end,
+rspamd_config:register_symbol({
+  name = 'DATE_IN_PAST',
   score = 1.0,
   description = 'Message date is in past',
   group = 'headers',
-  type = 'mime',
-}
+  type = 'virtual',
+  parent = date_id,
+})
 
 local obscured_id = rspamd_config:register_symbol{
   callback = function(task)
-    local urls = task:get_urls()
+    local susp_urls = task:get_urls_filtered({ 'obscured', 'zw_spaces'})
 
-    if urls then
-      for _,u in ipairs(urls) do
-        local fl = u:get_flags()
-        if fl.obscured then
+    if susp_urls and susp_urls[1] then
+      local obs_flag = url_flags_tab.obscured
+      local zw_flag = url_flags_tab.zw_spaces
+
+      for _,u in ipairs(susp_urls) do
+        local fl = u:get_flags_num()
+        if bit.band(fl, obs_flag) ~= 0 then
           task:insert_result('R_SUSPICIOUS_URL', 1.0, u:get_host())
         end
-        if fl.zw_spaces then
+        if bit.band(fl, zw_flag) ~= 0 then
           task:insert_result('ZERO_WIDTH_SPACE_URL', 1.0, u:get_host())
         end
       end
@@ -122,7 +148,7 @@ local obscured_id = rspamd_config:register_symbol{
   name = 'R_SUSPICIOUS_URL',
   score = 5.0,
   one_shot = true,
-  description = 'Obfusicated or suspicious URL has been found in a message',
+  description = 'Obfuscated or suspicious URL has been found in a message',
   group = 'url'
 }
 
@@ -345,14 +371,14 @@ rspamd_config.HAS_ATTACHMENT = {
 
 -- Requires freemail maps loaded in multimap
 local function freemail_reply_neq_from(task)
-  local frt = task:get_symbol('FREEMAIL_REPLYTO')
-  local ff  = task:get_symbol('FREEMAIL_FROM')
-  if (frt and ff and frt['options'] and ff['options'] and
-      frt['options'][1] ~= ff['options'][1])
-  then
-    return true
+  if not task:has_symbol('FREEMAIL_REPLYTO') or not task:has_symbol('FREEMAIL_FROM') then
+    return false
   end
-  return false
+  local frt = task:get_symbol('FREEMAIL_REPLYTO')
+  local ff = task:get_symbol('FREEMAIL_FROM')
+  local frt_opts = frt[1]['options']
+  local ff_opts = ff[1]['options']
+  return ( frt_opts and ff_opts and frt_opts[1] ~= ff_opts[1] )
 end
 
 rspamd_config:register_symbol({
@@ -379,7 +405,10 @@ rspamd_config.OMOGRAPH_URL = {
         if u:is_phished() then
 
           local h1 = u:get_host()
-          local h2 = u:get_phished():get_host()
+          local h2 = u:get_phished()
+          if h2 then -- Due to changes of the phished flag in 2.8
+            h2 = h2:get_host()
+          end
           if h1 and h2 then
             local selt = string.format('%s->%s', h1, h2)
             if not seen[selt] and util.is_utf_spoofed(h1, h2) then
@@ -483,7 +512,7 @@ local aliases_id = rspamd_config:register_symbol{
         end
 
         if modified then
-          task:set_recipients(type, addrs)
+          task:set_recipients(type, addrs, 'alias')
           task:insert_result('TAGGED_RCPT', 1.0, all_tags)
         end
       end
@@ -521,7 +550,7 @@ local check_from_display_name = rspamd_config:register_symbol{
     local from = task:get_from(2)
     if not (from and from[1] and from[1].name) then return false end
     -- See if we can parse an email address from the name
-    local parsed = util.parse_mail_address(from[1].name, task:get_mempool())
+    local parsed = rspamd_parsers.parse_mail_address(from[1].name, task:get_mempool())
     if not parsed then return false end
     if not (parsed[1] and parsed[1]['addr']) then return false end
     -- Make sure we did not mistake e.g. <something>@<name> for an email address
@@ -602,7 +631,7 @@ rspamd_config.SPOOF_REPLYTO = {
     end
     if not found_fromdom then return false end
     -- Parse Reply-To header
-    local parsed = ((util.parse_mail_address(rt, task:get_mempool()) or E)[1] or E).domain
+    local parsed = ((rspamd_parsers.parse_mail_address(rt, task:get_mempool()) or E)[1] or E).domain
     if not parsed then return false end
     -- Reply-To domain must be different to From domain
     if not util.strequal_caseless(parsed, from[1].domain) then
@@ -618,8 +647,9 @@ rspamd_config.SPOOF_REPLYTO = {
 
 rspamd_config.INFO_TO_INFO_LU = {
   callback = function(task)
-    local lu = task:get_header('List-Unsubscribe')
-    if not lu then return false end
+    if not task:has_header('List-Unsubscribe') then
+      return false
+    end
     local from = task:get_from('mime')
     if not (from and from[1] and util.strequal_caseless(from[1].user, 'info')) then
       return false
@@ -685,12 +715,14 @@ local check_encrypted_name = rspamd_config:register_symbol{
     local function check_part(part)
       if part:is_multipart() then
         local children = part:get_children() or {}
+        local text_kids = {}
 
         for _,cld in ipairs(children) do
           if cld:is_multipart() then
             check_part(cld)
           elseif cld:is_text() then
             seen_text = true
+            text_kids[#text_kids + 1] = cld
           else
             local type,subtype,_ = cld:get_type_full()
 
@@ -712,6 +744,17 @@ local check_encrypted_name = rspamd_config:register_symbol{
               end
             end
           end
+          if seen_text and seen_encrypted then
+            -- Ensure that our seen text is not really part of pgp #3205
+            for _,tp in ipairs(text_kids) do
+              local t,_ = tp:get_type()
+              seen_text = false -- reset temporary
+              if t and t == 'text' then
+                seen_text = true
+                break
+              end
+            end
+          end
         end
       end
     end
